/*** PIDgen.h *************************************************/
/*                                                            */
/* Functions to support a pseudonymization service            */
/*   PID = Pseudonymous IDentificator                         */
/*                                                            */
/* 1. Generate a PID from a current number: PIDgen            */
/*      The correspondence between ID data (or current        */
/*      number) and PID has to be stored elsewhere.           */
/*                                                            */
/* 2. Check a given PID for validity:       PIDcheck          */
/*      The algorithm recognizes up to 2 errors in the PID    */
/*      and corrects one error or a transposition of adjacent */
/*      characters.                                           */
/*------------------------------------------------------------*/
/* The encryption keys k1, k2, k3 and the number rndwidth of  */
/*   bits for randomization (global variables) must be set    */
/*   before applying PIDgen.                                  */
/* Important note: The keys should be different for different */
/*   application contexts; otherwise the pseudonymous         */
/*   property of the PID is weakened.                         */
/*------------------------------------------------------------*/
/* Klaus Pommerening, IMSD, Johannes-Gutenberg-Universitaet,  */
/*   Mainz, 3. April 2001                                     */
/* Version 1.00, 29. Mai 2004                                 */
/**************************************************************/

typedef unsigned long UINT30;         /* 30-bit-integers >= 0 */

UINT30 k1, k2, k3;            /* encryption keys              */
unsigned rndwidth;            /* bits for randomization       */
  /*----------------------------------------------------------*/
  /* A nonnegative integer of up to (30-rndwidth) bits gets   */
  /*   randomized by prefixing it with rndwidth random bits.  */
  /*                                                          */
  /*      | random part |  number      |                      */
  /* Bits: 29   ...    k  k-1   ...   0                       */
  /*      |< rndwidth  >|              |                      */
  /*                   1   0    ...   0  = rndfact = 2^k      */
  /*----------------------------------------------------------*/
  /* Constraints:  0 <= rndwith <= 12.                        */
  /* rndwidth = 0 implies *no randomization*.                 */
  /*----------------------------------------------------------*/

char *PIDgen(UINT30 x);
  /************************************************************/
  /* Input : Integer x, where 0 < x < 2^(30-rndwidth)         */
  /* Output: String of length 8 over the alphabet             */
  /*   Sigma = {0123456789ACDEFGHJKLMNPQRTUVWXYZ}             */
  /*   of 32 characters.                                      */
  /*   [BIOS left out because of similarity with 8105.]       */
  /* If the output is needed beyond the next invocation of    */
  /*   PIDgen, it must be copied to another variable.         */
  /*   PIDgen reuses the allocated memory.                    */
  /*----------------------------------------------------------*/
  /* The first 6 output characters consist of a cryptographic */
  /*   transformation of the randomized 30 bit input integer. */
  /*   Remark: The output space has 32^6 = 2^30 elements.     */
  /*   [The invalid input x = 0 maps to "000000".]            */
  /* The last 2 output characters consist of check bits.      */
  /* The mapping is injective: Distinct input numbers give    */
  /*   distinct PIDs.                                         */
  /*----------------------------------------------------------*/
  /* Error handling: If x is outside the required range,      */
  /*   the string "00000000" is returned.                     */
  /************************************************************/

int PIDcheck(char *p, char *a);
  /************************************************************/
  /* Input : Character string p                               */
  /* Result:  0, if the string is not a valid PID and at      */
  /*             least 2 characters should be changed to make */
  /*             it valid,                                    */
  /*          n  with 1 <= n <= 8, if the string is not a     */
  /*             valid PID, but becomes valid when the n-th   */
  /*             character is replaced with a,                */
  /*          n  with 9 <= n <= 15, if the string is not a    */
  /*             valid PID, but becomes valid when the        */
  /*             (n-8)-th and (n-7)-th character interchange. */
  /*         -1, if the string is a valid PID.                */
  /* Output variable: a - indicates possible error correction.*/
  /*             If the result is 0, or -1, or >= 9, then     */
  /*             the value of a is meaningless.               */
  /*----------------------------------------------------------*/
  /* The algorithm checks                                     */
  /* - whether the input string has length 8,                 */
  /* - whether the input string consists of valid characters, */
  /*     -- valid characters are as in Sigma above, or the    */
  /*        corresponding small letters; in other words, the  */
  /*        check is case insensitive --                      */
  /* - whether the last 2 characters consist of valid         */
  /*     check bits for the first 6 characters.               */
  /************************************************************/
